"""Functions to query GPUs in the system."""

import typing as t

from .available_features import cuda, GPU

from .errors import QueryError


def query_gpus(**_) -> t.List[t.Mapping[str, t.Any]]:
    """Get information about all GPUs."""
    if not GPU:
        return []
    gpus = []
    for i in range(cuda.Device.count()):
        device = cuda.Device(i)
        gpus.append(query_gpu(device))
    return gpus


def query_gpu(device: 'cuda.Device') -> t.Mapping[str, t.Any]:
    """Get information about a given GPU."""
    attributes = device.get_attributes()
    compute_capability = device.compute_capability()
    multiprocessors = attributes[cuda.device_attribute.MULTIPROCESSOR_COUNT]
    cuda_cores = calculate_cuda_cores(compute_capability, multiprocessors)
    try:
        return {
            'brand': device.name(),
            'memory': device.total_memory(),
            'memory_clock': attributes[cuda.device_attribute.MEMORY_CLOCK_RATE],
            'compute_capability': float('.'.join(str(_) for _ in compute_capability)),
            'clock': attributes[cuda.device_attribute.CLOCK_RATE],
            'multiprocessors': multiprocessors,
            'cores': cuda_cores,
            'warp_size': attributes[cuda.device_attribute.WARP_SIZE]
            }
    except KeyError as err:
        raise QueryError(
            f'expected value not present among device attributes: {device.get_attributes()}'
            ) from err


COMPUTE_CAPABILITY_MULTIPLIERS = {
    # Fermi
    2: {1: 48, None: 32},
    # Kepler
    3: 192,
    # Maxwell
    5: 128,
    # Pascal
    6: {0: 64, 1: 128}
}


def calculate_cuda_cores(compute_capability: t.Tuple[int, int],
                         multiprocessors: int) -> t.Optional[int]:
    """Calculate number of cuda cores according to Nvidia's specifications.

    Source: https://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#compute-capabilities
    """
    multiplier_major = COMPUTE_CAPABILITY_MULTIPLIERS.get(compute_capability[0])
    if isinstance(multiplier_major, int):
        return multiprocessors * multiplier_major
    if isinstance(multiplier_major, dict):
        multiplier_minor = multiplier_major.get(compute_capability[1])
        if multiplier_minor is None and None in multiplier_major:
            multiplier_minor = multiplier_major[None]
        if multiplier_minor is None:
            return None
        return multiprocessors * multiplier_minor
    return None



"""Functions to query system's RAM."""

import logging
import subprocess
import typing as t
from xml.etree import ElementTree as ET

from .available_features import psutil, RAM_TOTAL

_LOG = logging.getLogger(__name__)


def query_ram_total() -> t.Optional[int]:
    """Get information about total available RAM."""
    if not RAM_TOTAL:
        return None
    return psutil.virtual_memory().total


def query_ram(sudo: bool = False, **kwargs) -> t.Mapping[str, t.Any]:
    """Get all available information about RAM."""
    total_ram = query_ram_total()
    ram_banks = query_ram_banks(sudo=sudo, **kwargs)
    ram: t.Dict[str, t.Any] = {'total': total_ram}
    if ram_banks:
        ram['banks'] = ram_banks
    return ram


def query_ram_banks(sudo: bool = False, **_) -> t.List[t.Mapping[str, t.Any]]:
    """Extract information about RAM dice installed in the system."""
    try:
        xml_root = parse_lshw(sudo=sudo)
    except subprocess.TimeoutExpired:
        return []
    except subprocess.CalledProcessError:
        return []
    except FileNotFoundError:
        return []
    except ET.ParseError:
        return []
    nodes = xml_root.findall('.//node')
    _LOG.debug('%i nodes', len(nodes))
    ram_banks = []
    for node in nodes:
        node_id = node.attrib['id']
        _LOG.debug('%s', node_id)
        if not node_id.startswith('bank'):
            continue
        ram_banks.append(query_ram_bank(node))
    return ram_banks


def parse_lshw(sudo: bool = False):
    """Get RAM information via lshw."""
    cmd = (['sudo'] if sudo else []) + ['lshw', '-xml', '-quiet']
    result = subprocess.run(
        cmd, check=True, timeout=5, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return ET.fromstring(result.stdout.decode())


def query_ram_bank(node: ET.Element) -> t.Mapping[str, t.Any]:
    """Extract information about given RAM bank from XML node."""
    bank_size = node.findall('./size')
    bank_clock = node.findall('./clock')
    if len(bank_size) != 1 or len(bank_clock) != 1:
        _LOG.warning(
            'there should be exactly one size and clock value for a bank'
            ' but there are %i and %i respectively', len(bank_size), len(bank_clock))
    _LOG.debug(ET.tostring(node, encoding='utf8', method='xml').decode())
    assert bank_size[0].text is not None
    ram_bank = {'memory': int(bank_size[0].text)}
    try:
        bank_clock_text = bank_clock[0].text
    except IndexError:
        pass
    else:
        if bank_clock_text is not None:
            ram_bank['clock'] = int(bank_clock_text)
    return ram_bank



"""Tests for gpu_info module."""

import collections.abc
import unittest

from system_query.gpu_info import query_gpus, calculate_cuda_cores


class Tests(unittest.TestCase):

    def test_query_gpus(self):
        info = query_gpus()
        self.assertIsInstance(info, collections.abc.Sequence)

    def test_calculate_cuda_cores(self):
        for multiprocessors in (1, 2, 4, 128, 1024):
            self.assertEqual(calculate_cuda_cores((1, 0), multiprocessors), None)
            self.assertEqual(calculate_cuda_cores((2, 0), multiprocessors), 32 * multiprocessors)
            self.assertEqual(calculate_cuda_cores((2, 1), multiprocessors), 48 * multiprocessors)
            self.assertEqual(calculate_cuda_cores((2, 2), multiprocessors), 32 * multiprocessors)
            self.assertEqual(calculate_cuda_cores((5, 0), multiprocessors), 128 * multiprocessors)
            self.assertEqual(calculate_cuda_cores((5, 2), multiprocessors), 128 * multiprocessors)
            self.assertEqual(calculate_cuda_cores((5, 3), multiprocessors), 128 * multiprocessors)
            self.assertEqual(calculate_cuda_cores((6, 0), multiprocessors), 64 * multiprocessors)
            self.assertEqual(calculate_cuda_cores((6, 1), multiprocessors), 128 * multiprocessors)
            self.assertEqual(calculate_cuda_cores((6, 2), multiprocessors), None)
            self.assertEqual(calculate_cuda_cores((8, 0), multiprocessors), None)
            self.assertEqual(calculate_cuda_cores((9, 0), multiprocessors), None)



"""Tests for using system_query from command line."""

import ast
import contextlib
import io
import itertools
import json
import logging
import os
import tempfile
import unittest

from boilerplates.packaging_tests import run_module

_LOG = logging.getLogger(__name__)


@contextlib.contextmanager
def temporarily_set_logger_level(logger_name: str, level: int):
    """Change logger level on enter and restore on exit of this context."""
    logger = logging.getLogger(logger_name)
    level_ = logger.level
    logger.setLevel(level)
    try:
        yield
    finally:
        logger.setLevel(level_)


def preserve_logger_level(logger_name: str):
    return temporarily_set_logger_level(logger_name, logging.getLogger(logger_name).level)


class Tests(unittest.TestCase):

    def test_not_as_main(self):
        sio = io.StringIO()
        with contextlib.redirect_stdout(sio):
            run_module('system_query', run_name=None)
        self.assertEqual(sio.getvalue(), '')

    def test_help(self):
        sio = io.StringIO()
        with contextlib.redirect_stdout(sio):
            with preserve_logger_level('system_query'):
                with self.assertRaises(SystemExit):
                    run_module('system_query', '-h')
        _LOG.info('%s', sio.getvalue())

    def test_default_format(self):
        sio = io.StringIO()
        with contextlib.redirect_stdout(sio):
            with preserve_logger_level('system_query'):
                run_module('system_query')
        self.assertIsInstance(ast.literal_eval(sio.getvalue()), dict)

        for args in itertools.product(['--scope'], ['all', 'cpu', 'gpu', 'ram']):
            sio = io.StringIO()
            with contextlib.redirect_stdout(sio):
                with preserve_logger_level('system_query'):
                    run_module('system_query', *args)
            self.assertIsInstance(ast.literal_eval(sio.getvalue()), (list, dict), msg=args)

    def test_json(self):
        sio = io.StringIO()
        with contextlib.redirect_stdout(sio):
            with preserve_logger_level('system_query'):
                run_module('system_query', '--format', 'json')
        json_data = json.loads(sio.getvalue())
        self.assertIsNotNone(json_data)

        for args in itertools.product(['--scope'], ['all', 'cpu', 'gpu', 'ram']):
            sio = io.StringIO()
            with contextlib.redirect_stdout(sio):
                with preserve_logger_level('system_query'):
                    run_module('system_query', '--format', 'json', *args)
            json_data = json.loads(sio.getvalue())
            self.assertIsNotNone(json_data, msg=args)

    def test_raw(self):
        sio = io.StringIO()
        with contextlib.redirect_stdout(sio):
            with preserve_logger_level('system_query'):
                run_module('system_query', '--format', 'raw')
        self.assertIsInstance(ast.literal_eval(sio.getvalue()), dict)

    def test_file(self):
        loaders = {'raw': ast.literal_eval, 'json': json.loads}
        for fmt in ('raw', 'json'):
            with tempfile.NamedTemporaryFile(delete=False) as tmpfile:
                tmpfile_name = tmpfile.name
            with preserve_logger_level('system_query'):
                run_module('system_query', '--format', fmt, '--target', tmpfile_name)
            with open(tmpfile_name, encoding='utf-8') as tmpfile:
                data = loaders[fmt](tmpfile.read())
                self.assertIsInstance(data, dict)
                self.assertGreater(len(data), 0)
            os.remove(tmpfile_name)



"""Functions to query the operating system."""

import platform


def query_os():
    """Get information about OS."""
    # distro = distro.linux_distribution(full_distribution_name=False)
    system = platform.platform()  # + ' ' + ' '.join(str(_) for _ in distro)}
    return system



"""Tests for software_info module."""

import collections.abc
import unittest

from system_query.software_info import query_software


class Tests(unittest.TestCase):

    def test_query_software(self):
        info = query_software()
        self.assertIsInstance(info, collections.abc.Mapping)
